# coding=utf8
# ref: https://thrift.apache.org/docs/idl

from ply import yacc
from . import ThriftSyntaxError
from .lexer import tokens
from .model import *


def p_error(p):
    raise ThriftSyntaxError(p)


def p_start(p):
    '''start : header definition'''


def p_header(p):
    '''header : header_unit header
              |'''


def p_header_unit(p):
    '''header_unit : include
                   | namespace'''


def p_include(p):
    '''include : INCLUDE LITERAL'''
    thrift.includes.append(p[2])


def p_namespace(p):
    '''namespace : NAMESPACE namespace_scope IDENTIFIER'''
    thrift.namespaces[p[3]] = p[2]


def p_namespace_scope(p):
    '''namespace_scope : '*'
                       | IDENTIFIER'''
    p[0] = p[1]


def p_sep(p):
    '''sep : ','
           | ';'
    '''


def p_definition(p):
    '''definition : definition definition_unit
                  |'''


def p_definition_unit(p):
    '''definition_unit : const
                       | typedef
                       | enum
                       | struct
                       | union
                       | exception
                       | service
    '''


def p_const(p):
    '''const : CONST field_type IDENTIFIER '=' const_value '''
    thrift.consts[p[3]] = p[2].cast(p[5])


def p_const_value(p):
    '''const_value : INTCONSTANT
                   | DUBCONSTANT
                   | LITERAL
                   | BOOLCONSTANT
                   | IDENTIFIER
                   | const_list
                   | const_map'''

    p[0] = p[1]


def p_const_list(p):
    '''const_list : '[' const_value_seq ']' '''
    p[0] = p[2]


def p_const_value_seq(p):
    '''const_value_seq : const_value sep const_value_seq
                       | const_value const_value_seq
                       |'''
    _parse_seq(p)


def p_const_map(p):
    '''const_map : '{' const_map_seq '}' '''
    p[0] = dict(p[2])


def p_const_map_items(p):
    '''const_map_seq : const_map_item sep const_map_seq
                     | const_map_item const_map_seq
                     |
    '''
    _parse_seq(p)


def p_const_map_item(p):
    '''const_map_item : const_value ':'  const_value'''
    p[0] = [p[1], p[3]]


def p_typedef(p):
    '''typedef : TYPEDEF definition_type IDENTIFIER'''
    thrift.typedefs[p[3]] = p[2]


def p_enum(p):
    '''enum : ENUM IDENTIFIER '{' enum_seq '}' '''

    dct = dict(p[4])
    num = 0
    vals = dct.values()

    for key in dct:
        if dct[key] is None:
            while num in vals:
                num += 1
            dct[key] = num

    thrift.enums[p[2]] = dct


def p_enum_seq(p):
    '''enum_seq : enum_item sep enum_seq
                | enum_item enum_seq
                |
    '''
    _parse_seq(p)


def p_enum_item(p):
    '''enum_item : IDENTIFIER '=' INTCONSTANT
                 | IDENTIFIER'''
    if len(p) == 4:
        p[0] = [p[1], p[3]]
    elif len(p) == 2:
        p[0] = [p[1], None]


def p_struct(p):
    '''struct : STRUCT IDENTIFIER '{' field_seq '}' '''
    thrift.structs[p[2]] = p[4]


def p_union(p):
    '''union : UNION IDENTIFIER '{' field_seq '}' '''
    thrift.unions[p[2]] = p[4]


def p_exception(p):
    '''exception : EXCEPTION IDENTIFIER '{' field_seq '}' '''
    thrift.exceptions[p[2]] = p[4]


def p_service(p):
    '''service : SERVICE IDENTIFIER '{' function_seq '}'
               | SERVICE IDENTIFIER EXTENDS IDENTIFIER '{' function_seq '}'
    '''
    apis = {}
    extends = p[4] if len(p) == 8 else None
    functions = p[len(p) - 2]

    for function in functions:
        apis[function.name] = function

    thrift.services[p[2]] = Service(extends=extends, apis=apis)


def p_function(p):
    '''function : ONEWAY function_type IDENTIFIER '(' field_seq ')' throws
                | ONEWAY function_type IDENTIFIER '(' field_seq ')'
                | function_type IDENTIFIER '(' field_seq ')' throws
                | function_type IDENTIFIER '(' field_seq ')'
    '''

    if len(p) == 8:
        p[0] = Function(p[2], p[3], fields=p[5], throws=p[7], oneway=True)
    elif len(p) == 7 and p[1] == 'oneway':
        p[0] = Function(p[2], p[3], fields=p[5], throws=None, oneway=True)
    elif len(p) == 7 and p[1] != 'oneway':
        p[0] = Function(p[1], p[2], fields=p[4], throws=p[6], oneway=False)
    elif len(p) == 6:
        p[0] = Function(p[1], p[2], fields=p[4], throws=None, oneway=False)


def p_function_seq(p):
    '''function_seq : function sep function_seq
                    | function function_seq
                    |
    '''
    _parse_seq(p)


def p_throws(p):
    '''throws : THROWS '(' field_seq ')' '''
    p[0] = p[3]


def p_function_type(p):
    '''function_type : field_type
                     | VOID'''
    p[0] = p[1]


def p_field_seq(p):
    '''field_seq : field sep field_seq
                 | field field_seq
                 |
    '''
    _parse_seq(p)


def p_field(p):
    '''field : field_id field_req field_type IDENTIFIER
             | field_id field_req field_type IDENTIFIER '=' const_value'''
    v = p[6] if len(p) == 7 else None
    p[0] = Field(p[1], p[3], p[4], value=v, requirement=p[2])


def p_field_id(p):
    '''field_id : INTCONSTANT ':' '''
    p[0] = p[1]


def p_field_req(p):
    '''field_req : REQUIRED
                 | OPTIONAL
                 |'''
    if len(p) == 2:
        p[0] = p[1]


def p_field_type(p):
    '''field_type : IDENTIFIER
                  | base_type
                  | container_type'''
    p[0] = p[1]


def p_base_type(p):
    '''base_type : BOOL
                 | BYTE
                 | I16
                 | I32
                 | I64
                 | DOUBLE
                 | STRING
                 | BINARY'''

    p[0] = BASE_TYPE_MAPS[p[1]]()


def p_container_type(p):
    '''container_type : map_type
                      | list_type
                      | set_type'''
    p[0] = p[1]


def p_map_type(p):
    '''map_type : MAP '<' field_type ',' field_type '>' '''
    p[0] = MapType(['map', p[3], p[5]])


def p_list_type(p):
    '''list_type : LIST '<' field_type '>' '''
    p[0] = ListType(['list', p[3]])


def p_set_type(p):
    '''set_type : SET '<' field_type '>' '''
    p[0] = SetType(['set', p[3]])


def p_definition_type(p):
    '''definition_type : base_type
                       | container_type'''
    p[0] = p[1]


parser = yacc.yacc(debug=False, write_tables=0)

thrift = None


def parse(data):
    global thrift
    thrift = Thrift()
    parser.parse(data)
    return thrift


def _parse_seq(p):
    if len(p) == 4:
        p[0] = [p[1]] + p[3]
    elif len(p) == 3:
        p[0] = [p[1]] + p[2]
    elif len(p) == 1:
        p[0] = []



# coding=utf8

import sys

if sys.version_info[0] < 3:
    bytes_ = bytes
else:
    bytes_ = lambda x: bytes(x, 'utf8')


class Thrift(dict):

    def __init__(self,
                 includes=None,
                 namespaces=None,
                 consts=None,
                 enums=None,
                 typedefs=None,
                 structs=None,
                 unions=None,
                 exceptions=None,
                 services=None):
        super(Thrift, self).__init__()

        if includes is None:
            includes = []
        if namespaces is None:
            namespaces = {}
        if consts is None:
            consts = {}
        if enums is None:
            enums = {}
        if typedefs is None:
            typedefs = {}
        if structs is None:
            structs = {}
        if unions is None:
            unions = {}
        if exceptions is None:
            exceptions = {}
        if services is None:
            services = {}

        self.includes = self['includes'] = includes
        self.namespaces = self['namespaces'] = namespaces
        self.consts = self['consts'] = consts
        self.enums = self['enums'] = enums
        self.typedefs = self['typedefs'] = typedefs
        self.structs = self['structs'] = structs
        self.unions = self['unions'] = unions
        self.exceptions = self['exceptions'] = exceptions
        self.services = self['services'] = services


class BaseType(str):
    pass


class BoolType(BaseType):

    cast = bool

    def __new__(self):
        return str.__new__(self, 'bool')


class ByteType(BaseType):

    cast = int

    def __new__(self):
        return str.__new__(self, 'byte')


class I16Type(BaseType):

    cast = int

    def __new__(self):
        return str.__new__(self, 'i16')


class I32Type(BaseType):

    cast = int

    def __new__(self):
        return str.__new__(self, 'i32')


class I64Type(BaseType):

    cast = int

    def __new__(self):
        return str.__new__(self, 'i64')


class DoubleType(BaseType):

    cast = float

    def __new__(self):
        return str.__new__(self, 'double')


class StringType(BaseType):

    cast = str

    def __new__(self):
        return str.__new__(self, 'string')


class BinaryType(BaseType):

    cast = bytes_

    def __new__(self):
        return str.__new__(self, 'binary')


class ContainerType(list):
    pass


class ListType(ContainerType):   # ['list', val_type]

    def cast(self, data):
        return map(self[1].cast, data)


class MapType(ContainerType):  # ['map', k_type, v_type]

    def cast(self, data):
        dct = {}
        keys = data.keys()

        for key in keys:
            dct[self[1].cast(key)] = self[2].cast(data[key])
        return dct


class SetType(ContainerType):  # ['set', v_type]
    pass


BASE_TYPE_MAPS = {
    'bool': BoolType,
    'byte': ByteType,
    'i16': I16Type,
    'i32': I32Type,
    'i64': I64Type,
    'double': DoubleType,
    'string': StringType,
    'binary': BinaryType
}


class Field(dict):

    def __init__(self, id, type, name, value=None, requirement=None):
        super(Field, self).__init__()
        self.id = self['id'] = id
        self.type = self['type'] = type
        self.name = self['name'] = name
        self.value = self['value'] = value
        self.requirement = self['requirement'] = requirement


class Function(dict):

    def __init__(self, type, name, fields=None, throws=None, oneway=False):
        super(Function, self).__init__()

        if fields is None:
            fields = []
        if throws is None:
            throws = []
        self.type = self['type'] = type
        self.name = self['name'] = name
        self.fields = self['fields'] = fields
        self.throws = self['throws'] = throws
        self.oneway = self['oneway'] = oneway


class Service(dict):

    def __init__(self, extends=None, apis=None):
        super(Service, self).__init__()

        if apis is None:
            apis = {}

        self.extends = self['extends'] = extends
        self.apis = self['apis'] = apis



# coding=utf8

"""
thriftr
~~~~~~~

The pure python parser for apache thrift.

Author: hit9 <wangchao@ele.me>
License: MIT.

    >>> from thriftr import parse
    >>> thrift = parse(open('sample.thrift').read())
"""

__version__ = '0.0.1'


class ThriftSyntaxError(Exception):
    pass


from .parser import parse



# coding=utf8

from ply import lex

from . import ThriftSyntaxError


literals = ':;,=*{}()<>[]'


keywords = (
    'namespace',
    'include',
    'void',
    'bool',
    'byte',
    'i16',
    'i32',
    'i64',
    'double',
    'string',
    'binary',
    'map',
    'list',
    'set',
    'oneway',
    'typedef',
    'struct',
    'union',
    'exception',
    'extends',
    'throws',
    'service',
    'enum',
    'const',
    'required',
    'optional',
)


tokens = (
    'BOOLCONSTANT',
    'INTCONSTANT',
    'DUBCONSTANT',
    'LITERAL',
    'IDENTIFIER',
) + tuple(map(lambda kw: kw.upper(), keywords))


t_ignore = ' \t\r'   # whitespace


def t_error(t):
    raise ThriftSyntaxError('Illegal characher %r at line %d' %
                            (t.value[0], t.lineno))


def t_newline(t):
    r'\n+'
    t.lexer.lineno += len(t.value)


def t_ignore_SILLYCOMM(t):
    r'\/\*\**\*\/'
    t.lexer.lineno += t.value.count('\n')


def t_ignore_MULTICOMM(t):
    r'\/\*[^*]\/*([^*/]|[^*]\/|\*[^/])*\**\*\/'
    t.lexer.lineno += t.value.count('\n')


def t_ignore_DOCTEXT(t):
    r'\/\*\*([^*/]|[^*]\/|\*[^/])*\**\*\/'
    t.lexer.lineno += t.value.count('\n')


def t_ignore_UNIXCOMMENT(t):
    r'\#[^\n]*'


def t_ignore_COMMENT(t):
    r'\/\/[^\n]*'


def t_BOOLCONSTANT(t):
    r'true|false'
    t.value = t.value == 'true'
    return t


def t_DUBCONSTANT(t):
    r'-?\d+\.\d*(e-?\d+)?'
    t.value = float(t.value)
    return t


def t_HEXCONSTANT(t):
    r'0x[0-9A-Fa-f]+'
    t.value = int(t.value, 16)
    t.type = 'INTCONSTANT'
    return t


def t_INTCONSTANT(t):
    r'[+-]?[0-9]+'
    t.value = int(t.value)
    return t


def t_LITERAL(t):
    r'(\"([^\\\n]|(\\.))*?\")|\'([^\\\n]|(\\.))*?\''
    s = t.value[1:-1]
    maps = {
        't': '\t',
        'r': '\r',
        'n': '\n',
        '\\': '\\',
        '\'': '\'',
        '"': '\"'
    }
    i = 0
    length = len(s)
    val = ''
    while i < length:
        if s[i] == '\\':
            i += 1
            if s[i] in maps:
                val += maps[s[i]]
            else:
                msg = 'Unexcepted escaping characher: %s' % s[i]
                raise ThriftSyntaxError(msg)
        else:
            val += s[i]

        i += 1

    t.value = val
    return t


def t_IDENTIFIER(t):
    r'[a-zA-Z_](\.[a-zA-Z_0-9]|[a-zA-Z_0-9])*'

    if t.value in keywords:
        t.type = t.value.upper()
        return t
    return t


lexer = lex.lex()



# coding=utf8

from setuptools import setup

setup(
    name='thriftr',
    version='0.0.1',
    author='hit9',
    author_email='nz2324@126.com',
    description='Pure python apache thrift parser',
    packages=['thriftr'],
    url='https://github.com/hit9/thriftr.git',
    license='MIT',
    install_requires=['ply'],
    zip_safe=False,
    classifiers=[
        "Topic :: Software Development",
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ]
)
